# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst
#
# Astropy documentation build configuration file.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this file.
#
# All configuration values have a default. Some values are defined in
# the global Astropy configuration which is loaded here before anything else.
# See astropy.sphinx.conf for which values are set there.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('..'))
# IMPORTANT: the above commented section was generated by sphinx-quickstart, but
# is *NOT* appropriate for astropy or Astropy affiliated packages. It is left
# commented out with this explanation to make it clear why this should not be
# done. If the sys.path entry above is added, when the astropy.sphinx.conf
# import occurs, it will import the *source* version of astropy instead of the
# version installed (if invoked as "make html" or directly with sphinx), or the
# version in the build directory (if "python setup.py build_sphinx" is used).
# Thus, any C-extensions that are needed to build the documentation will *not*
# be accessible, and the documentation will not build correctly.

import datetime
import os
import sys
from pathlib import Path


# Get configuration information from pyproject.toml
# try:
#     import tomllib
# except ImportError:
#     import toml as tomllib
#
# sys.path.insert(1, '..')
#
# setup_cfg = tomllib.load(
#     path.join(path.dirname(__file__), '..', 'pyproject.toml')
# )['project']

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

#sys.path.insert(0, os.path.abspath('../'))


def find_mod_objs_patched(*args, **kwargs):
    from sphinx_automodapi.utils import find_mod_objs

    return find_mod_objs(args[0], onlylocals=True)


def patch_automodapi(app):
    """Monkey-patch the automodapi extension to exclude imported members"""
    from sphinx_automodapi import automodsumm

    automodsumm.find_mod_objs = find_mod_objs_patched


def setup(app):
    app.connect("builder-inited", patch_automodapi)


with open(Path(__file__).parent.parent / "pyproject.toml", "rb") as configuration_file:
    conf = tomllib.load(configuration_file)
metadata = conf["project"]


extensions = [
    'sphinx_automodapi.automodapi',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
]
# -- General configuration ----------------------------------------------------

master_doc = 'index'

# If your documentation needs a minimal Sphinx version, state it here.
#needs_sphinx = '1.2'

# To perform a Sphinx version check that needs to be more specific than
# major.minor, call `check_sphinx_version("x.y.z")` here.
# check_sphinx_version("1.2.1")

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_templates']

# This is added to the end of RST files - a good place to put substitutions to
# be used globally.
rst_epilog = """
"""

# -- Project information ------------------------------------------------------

# This does not *have* to match the package name, but typically does
project = metadata['name']
author = metadata['authors'][0]['name']
copyright = f'{datetime.datetime.now().year}, {author }'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.

__import__(project)
package = sys.modules[project]

# The short X.Y version.
version = package.__version__.split('-', 1)[0]
# The full version, including alpha/beta/rc tags.
release = package.__version__


# -- Options for HTML output ---------------------------------------------------

# A NOTE ON HTML THEMES
# The global astropy configuration uses a custom theme, 'bootstrap-astropy',
# which is installed along with astropy. A different theme can be used or
# the options for this theme can be modified by overriding some of the
# variables set in the global configuration. The variables set in the
# global configuration are listed below, commented out.

# Add any paths that contain custom themes here, relative to this directory.
# To use a different custom theme, add the directory containing the theme.
#html_theme_path = []

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes. To override the custom theme, set this to the
# name of a builtin theme or the name of a custom theme in html_theme_path.
#html_theme = None

# Custom sidebar templates, maps document names to template names.
#html_sidebars = {}

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
#html_favicon = ''

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
#html_last_updated_fmt = ''

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = f'{project} v{release}'

# Output file base name for HTML help builder.
htmlhelp_basename = project + 'doc'


# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [('index', project + '.tex', project + u' Documentation',
                    author, 'manual')]


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [('index', project.lower(), project + u' Documentation',
              [author], 1)]



"""
STScI Python compatable blot module
"""
import numpy as np

from drizzle import calc_pixmap
from drizzle import cdrizzle


def doblot(source, source_wcs, blot_wcs, exptime, coeffs=True,
           interp='poly5', sinscl=1.0, stepsize=10, wcsmap=None):
    """
    Low level routine for performing the 'blot' operation.

    Create a single blotted image from a single source image. The
    interface is compatible with STScI code. All distortion information
    is assumed to be included in the WCS specification of the 'output'
    blotted image given in 'blot_wcs'.

    Parameters
    ----------

    source : 2d array
        Input numpy array of the source image in units of 'cps'.

    source_wcs : wcs
        The source image WCS.

    blot_wcs : wcs
        The blotted image WCS. The WCS that the source image will be
        resampled to.

    exptime : float
        The exposure time of the input image.

    interp : str, optional
        The type of interpolation used in the blotting. The
        possible values are "nearest" (nearest neighbor interpolation),
        "linear" (bilinear interpolation), "poly3" (cubic polynomial
        interpolation), "poly5" (quintic polynomial interpolation),
        "sinc" (sinc interpolation), "lan3" (3rd order Lanczos
        interpolation), and "lan5" (5th order Lanczos interpolation).

    sincscl : float, optional
        The scaling factor for sinc interpolation.

    Returns
    -------

    A 2d numpy array with the blotted image

    Other Parameters
    ----------------

    coeffs : bool, optional
        Not used. Only kept for backwards compatibility.

    stepsize : float, optional
        Was used when input to output mapping was computed
        internally. Is no longer used and only here for backwards compatibility.

    wcsmap : function, optional
        Was used when input to output mapping was computed
        internally. Is no longer used and only here for backwards compatibility.
    """
    _outsci = np.zeros(blot_wcs.pixel_shape[::-1], dtype=np.float32)

    # compute the undistorted 'natural' plate scale
    blot_wcs.sip = None
    blot_wcs.cpdis1 = None
    blot_wcs.cpdis2 = None
    blot_wcs.det2im = None

    pixmap = calc_pixmap.calc_pixmap(blot_wcs, source_wcs)
    pix_ratio = source_wcs.pscale / blot_wcs.pscale

    cdrizzle.tblot(source, pixmap, _outsci, scale=pix_ratio, kscale=1.0,
                   interp=interp, exptime=exptime, misval=0.0, sinscl=sinscl)

    return _outsci



import numpy as np


def find_keyword_extn(fimg, keyword, value=None):
    """
    This function will return the index of the extension in a multi-extension
    FITS file which contains the desired keyword with the given value.

    Parameters
    ----------

    fimg : hdulist
        A list of header data units

    keyword : str
        The keyword to search for

    value : str or number, optional
        If set, the value the keyword must have to match

    Returns
    -------

    The index of the extension
    """

    i = 0
    extnum = -1
    # Search through all the extensions in the FITS object
    for chip in fimg:
        hdr = chip.header
        # Check to make sure the extension has the given keyword
        if keyword in hdr:
            if value is not None:
                # If it does, then does the value match the desired value
                # MUST use 'str.strip' to match against any input string!
                if hdr[keyword].strip() == value:
                    extnum = i
                    break
            else:
                extnum = i
                break
        i += 1
    # Return the index of the extension which contained the
    # desired EXTNAME value.
    return extnum


def get_extn(fimg, extn=''):
    """
    Returns the FITS extension corresponding to extension specified in
    filename. Defaults to returning the first extension with data or the
    primary extension, if none have data.

    Parameters
    ----------

    fimg : hdulist
        A list of header data units

    extn : str
        The extension name and version to match

    Returns
    -------

    The matching header data unit
    """

    if extn:
        try:
            _extn = parse_extn(extn)
            if _extn[0]:
                _e = fimg.index_of(_extn)
            else:
                _e = _extn[1]

        except KeyError:
            _e = None

        if _e is None:
            _extn = None
        else:
            _extn = fimg[_e]

    else:
        # If no extension is provided, search for first extension
        # in FITS file with data associated with it.

        # Set up default to point to PRIMARY extension.
        _extn = fimg[0]
        # then look for first extension with data.
        for _e in fimg:
            if _e.data is not None:
                _extn = _e
                break

    return _extn


def get_keyword(fimg, keyword, default=None):
    """
    Return a keyword value from the header of an image,
    or the default if the keyword is not found.

    Parameters
    ----------

    fimg : hdulist
        A list of header data units

    keyword : hdulist
        The keyword value to search for

    default : str or number, optional
        The default value if not found

    Returns
    -------

    The value if found or default if not
    """

    value = None
    if keyword:
        _nextn = find_keyword_extn(fimg, keyword)
        try:
            value = fimg[_nextn].header[keyword]
        except KeyError:
            value = None

    if value is None and default is not None:
        value = default

    return value


def is_blank(value):
    """
    Determines whether or not a value is considered 'blank'.

    Parameters
    ----------

    value : str
        The value to check

    Returns
    -------

    True or False
    """
    return value.strip() == ""


def parse_extn(extn=''):
    """
    Parse a string representing a qualified fits extension name as in the
    output of parse_filename and return a tuple (str(extname), int(extver)),
    which can be passed to pyfits functions using the 'ext' kw.
    Default return is the first extension in a fits file.

    Examples
    --------
        >>> parse_extn('sci,2')
        ('sci', 2)
        >>> parse_extn('2')
        ('', 2)
        >>> parse_extn('sci')
        ('sci', 1)

    Parameters
    ----------
    extn : str
        The extension name

    Returns
    -------
    A tuple of the extension name and value
    """
    if not extn:
        return ('', 0)

    try:
        lext = extn.split(',')
    except Exception:
        return ('', 1)

    if len(lext) == 1 and lext[0].isdigit():
        return ("", int(lext[0]))
    elif len(lext) == 2:
        return (lext[0], int(lext[1]))
    else:
        return (lext[0], 1)


def parse_filename(filename):
    """
    Parse out filename from any specified extensions.
    Returns rootname and string version of extension name.

    Parameters
    ----------

    filename : str
        The filename to be parsed

    Returns
    -------

    A tuple with the filename root and extension
    """
    # Parse out any extension specified in filename
    _indx = filename.find('[')
    if _indx > 0:
        # Read extension name provided
        _fname = filename[:_indx]
        _extn = filename[_indx + 1:-1]
    else:
        _fname = filename
        _extn = ''

    return _fname, _extn


def set_pscale(the_wcs):
    """
    Calculates the plate scale from cdelt and the pc  matrix and adds
    it to the WCS. Plate scale is not part of the WCS standard, but is
    required by the drizzle code

    Parameters
    ----------

    the_wcs : wcs
        A WCS object
    """
    try:
        cdelt = the_wcs.wcs.get_cdelt()
        pc = the_wcs.wcs.get_pc()

    except Exception:
        try:
            # for non-celestial axes, get_cdelt doesnt work
            cdelt = the_wcs.wcs.cd * the_wcs.wcs.cdelt
        except AttributeError:
            cdelt = the_wcs.wcs.cdelt

        try:
            pc = the_wcs.wcs.pc
        except AttributeError:
            pc = 1

    pccd = np.array(cdelt * pc)
    scales = np.sqrt((pccd ** 2).sum(axis=0, dtype=float))
    the_wcs.pscale = scales[0]



"""
A package for combining dithered images into a single image
"""


from pkg_resources import get_distribution, DistributionNotFound
try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    __version__ = 'unknown'



import math
import os
import pytest

import numpy as np
from astropy import wcs
from astropy.io import fits

from drizzle import drizzle, cdrizzle


TEST_DIR = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.join(TEST_DIR, 'data')


ok = False


def bound_image(image):
    """
    Compute region where image is non-zero
    """
    coords = np.nonzero(image)
    ymin = coords[0].min()
    ymax = coords[0].max()
    xmin = coords[1].min()
    xmax = coords[1].max()
    return (ymin, ymax, xmin, xmax)


def centroid(image, size, center):
    """
    Compute the centroid of a rectangular area
    """
    ylo = int(center[0] - size / 2)
    yhi = min(ylo + size, image.shape[0])
    xlo = int(center[1] - size / 2)
    xhi = min(xlo + size, image.shape[1])

    center = [0.0, 0.0, 0.0]
    for y in range(ylo, yhi):
        for x in range(xlo, xhi):
            center[0] += y * image[y,x]
            center[1] += x * image[y,x]
            center[2] += image[y,x]

    if center[2] == 0.0:
        return None

    center[0] /= center[2]
    center[1] /= center[2]
    return center


def centroid_close(list_of_centroids, size, point):
    """
    Find if any centroid is close to a point
    """
    for i in range(len(list_of_centroids) - 1, -1, -1):
        if (abs(list_of_centroids[i][0] - point[0]) < int(size / 2) and
                abs(list_of_centroids[i][1] - point[1]) < int(size / 2)):
            return 1

    return 0


def centroid_compare(centroid):
    return centroid[1]


def centroid_distances(image1, image2, amp, size):
    """
    Compute a list of centroids and the distances between them in two images
    """
    distances = []
    list_of_centroids = centroid_list(image2, amp, size)
    for center2 in list_of_centroids:
        center1 = centroid(image1, size, center2)
        if center1 is None:
            continue

        disty = center2[0] - center1[0]
        distx = center2[1] - center1[1]
        dist = math.sqrt(disty * disty + distx * distx)
        dflux = abs(center2[2] - center1[2])
        distances.append([dist, dflux, center1, center2])

    distances.sort(key=centroid_compare)
    return distances


def centroid_list(image, amp, size):
    """
    Find the next centroid
    """
    list_of_centroids = []
    points = np.transpose(np.nonzero(image > amp))
    for point in points:
        if not centroid_close(list_of_centroids, size, point):
            center = centroid(image, size, point)
            list_of_centroids.append(center)

    return list_of_centroids


def centroid_statistics(title, fname, image1, image2, amp, size):
    """
    write centroid statistics to compare differences btw two images
    """
    stats = ("minimum", "median", "maximum")
    images = (None, None, image1, image2)
    im_type = ("", "", "test", "reference")

    diff = []
    distances = centroid_distances(image1, image2, amp, size)
    indexes = (0, int(len(distances) / 2), len(distances) - 1)
    fd = open(fname, 'w')
    fd.write("*** %s ***\n" % title)

    if len(distances) == 0:
        diff = [0.0, 0.0, 0.0]
        fd.write("No matches!!\n")

    elif len(distances) == 1:
        diff = [distances[0][0], distances[0][0], distances[0][0]]

        fd.write("1 match\n")
        fd.write("distance = %f flux difference = %f\n" %
                 (distances[0][0], distances[0][1]))

        for j in range(2, 4):
            ylo = int(distances[0][j][0]) - 1
            yhi = int(distances[0][j][0]) + 2
            xlo = int(distances[0][j][1]) - 1
            xhi = int(distances[0][j][1]) + 2
            subimage = images[j][ylo:yhi,xlo:xhi]
            fd.write("\n%s image centroid = (%f,%f) image flux = %f\n" %
                     (im_type[j], distances[0][j][0], distances[0][j][1],
                      distances[0][j][2]))
            fd.write(str(subimage) + "\n")

    else:
        fd.write("%d matches\n" % len(distances))

        for k in range(0,3):
            i = indexes[k]
            diff.append(distances[i][0])
            fd.write("\n%s distance = %f flux difference = %f\n" %
                     (stats[k],distances[i][0], distances[i][1]))

            for j in range(2, 4):
                ylo = int(distances[i][j][0]) - 1
                yhi = int(distances[i][j][0]) + 2
                xlo = int(distances[i][j][1]) - 1
                xhi = int(distances[i][j][1]) + 2
                subimage = images[j][ylo:yhi,xlo:xhi]
                fd.write("\n%s %s image centroid = (%f,%f) image flux = %f\n" %
                         (stats[k], im_type[j], distances[i][j][0],
                          distances[i][j][1], distances[i][j][2]))
                fd.write(str(subimage) + "\n")

    fd.close()
    return tuple(diff)


def make_point_image(input_image, point, value):
    """
    Create an image with a single point set
    """
    output_image = np.zeros(input_image.shape, dtype=input_image.dtype)
    output_image[point] = value
    return output_image


def make_grid_image(input_image, spacing, value):
    """
    Create an image with points on a grid set
    """
    output_image = np.zeros(input_image.shape, dtype=input_image.dtype)

    shape = output_image.shape
    half_space = int(spacing / 2)
    for y in range(half_space, shape[0], spacing):
        for x in range(half_space, shape[1], spacing):
            output_image[y,x] = value

    return output_image


def print_wcs(title, wcs):
    """
    Print the wcs header cards
    """
    print("=== %s ===" % title)
    print(wcs.to_header_string())


def read_image(filename):
    """
    Read the image from a fits file
    """
    path = os.path.join(DATA_DIR, filename)
    hdu = fits.open(path)

    image = hdu[1].data
    hdu.close()
    return image


def read_wcs(filename):
    """
    Read the wcs of a fits file
    """
    path = os.path.join(DATA_DIR, filename)
    hdu = fits.open(path)
    the_wcs = wcs.WCS(hdu[1].header)
    hdu.close()
    return the_wcs


def test_square_with_point(tmpdir):
    """
    Test do_driz square kernel with point
    """
    output = str(tmpdir.join('output_square_point.fits'))
    output_difference = str(tmpdir.join('difference_square_point.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_square_point.fits')

    insci = read_image(input_file)
    inwcs = read_wcs(input_file)
    insci = make_point_image(insci, (500, 200), 100.0)
    inwht = np.ones(insci.shape,dtype=insci.dtype)
    output_wcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="")
    driz.add_image(insci, inwcs, inwht=inwht)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("square with point",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 8)

        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


@pytest.mark.parametrize(
    'kernel', ['square', 'point', 'turbo', 'gaussian', 'lanczos3'],
)
def test_zero_input_weight(kernel):
    """
    Test do_driz square kernel with grid
    """
    # initialize input:
    insci = np.ones((200, 400), dtype=np.float32)
    inwht = np.ones((200, 400), dtype=np.float32)
    inwht[:, 150:155] = 0

    # initialize output:
    outsci = np.zeros((210, 410), dtype=np.float32)
    outwht = np.zeros((210, 410), dtype=np.float32)
    outctx = np.zeros((210, 410), dtype=np.int32)

    # define coordinate mapping:
    pixmap = np.moveaxis(np.mgrid[1:201, 1:401][::-1], 0, -1)

    # resample:
    cdrizzle.tdriz(
        insci, inwht, pixmap,
        outsci, outwht, outctx,
        uniqid=1,
        xmin=0, xmax=400,
        ymin=0, ymax=200,
        pixfrac=1,
        kernel=kernel,
        in_units='cps',
        expscale=1,
        wtscale=1,
        fillstr='INDEF',
    )

    # check that no pixel with 0 weight has any counts:
    assert np.sum(np.abs(outsci[(outwht == 0)])) == 0.0


def test_square_with_grid(tmpdir):
    """
    Test do_driz square kernel with grid
    """
    output = str(tmpdir.join('output_square_grid.fits'))
    output_difference = str(tmpdir.join('difference_square_grid.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_square_grid.fits')

    insci = read_image(input_file)
    inwcs = read_wcs(input_file)
    insci = make_grid_image(insci, 64, 100.0)
    inwht = np.ones(insci.shape,dtype=insci.dtype)
    output_wcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="")
    driz.add_image(insci, inwcs, inwht=inwht)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("square with grid",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 8)
        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_turbo_with_grid(tmpdir):
    """
    Test do_driz turbo kernel with grid
    """
    output = str(tmpdir.join('output_turbo_grid.fits'))
    output_difference = str(tmpdir.join('difference_turbo_grid.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_turbo_grid.fits')

    insci = read_image(input_file)
    inwcs = read_wcs(input_file)
    insci = make_grid_image(insci, 64, 100.0)
    inwht = np.ones(insci.shape,dtype=insci.dtype)
    output_wcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="", kernel='turbo')
    driz.add_image(insci, inwcs, inwht=inwht)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("turbo with grid",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 8)

        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_gaussian_with_grid(tmpdir):
    """
    Test do_driz gaussian kernel with grid
    """
    output = str(tmpdir.join('output_gaussian_grid.fits'))
    output_difference = str(tmpdir.join('difference_gaussian_grid.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_gaussian_grid.fits')

    insci = read_image(input_file)
    inwcs = read_wcs(input_file)
    insci = make_grid_image(insci, 64, 100.0)
    inwht = np.ones(insci.shape,dtype=insci.dtype)
    output_wcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="", kernel='gaussian')
    driz.add_image(insci, inwcs, inwht=inwht)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("gaussian with grid",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 8)

        assert med_diff < 1.0e-6
        assert max_diff < 2.0e-5


def test_lanczos_with_grid(tmpdir):
    """
    Test do_driz lanczos kernel with grid
    """
    output = str(tmpdir.join('output_lanczos_grid.fits'))
    output_difference = str(tmpdir.join('difference_lanczos_grid.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_lanczos_grid.fits')

    insci = read_image(input_file)
    inwcs = read_wcs(input_file)
    insci = make_grid_image(insci, 64, 100.0)
    inwht = np.ones(insci.shape,dtype=insci.dtype)
    output_wcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="", kernel='lanczos3')
    driz.add_image(insci, inwcs, inwht=inwht)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("lanczos with grid",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 8)
        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_tophat_with_grid(tmpdir):
    """
    Test do_driz tophat kernel with grid
    """
    output = str(tmpdir.join('output_tophat_grid.fits'))
    output_difference = str(tmpdir.join('difference_tophat_grid.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_tophat_grid.fits')

    insci = read_image(input_file)
    inwcs = read_wcs(input_file)
    insci = make_grid_image(insci, 64, 100.0)
    inwht = np.ones(insci.shape,dtype=insci.dtype)
    output_wcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="", kernel='tophat')
    driz.add_image(insci, inwcs, inwht=inwht)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("tophat with grid",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 8)
        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_point_with_grid(tmpdir):
    """
    Test do_driz point kernel with grid
    """
    output = str(tmpdir.join('output_point_grid.fits'))
    output_difference = str(tmpdir.join('difference_point_grid.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_point_grid.fits')

    insci = read_image(input_file)
    inwcs = read_wcs(input_file)
    insci = make_grid_image(insci, 64, 100.0)
    inwht = np.ones(insci.shape,dtype=insci.dtype)
    output_wcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="", kernel='point')
    driz.add_image(insci, inwcs, inwht=inwht)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("point with grid",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 8)
        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_blot_with_point(tmpdir):
    """
    Test do_blot with point image
    """
    output = str(tmpdir.join('output_blot_point.fits'))
    output_difference = str(tmpdir.join('difference_blot_point.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_blot_point.fits')

    outsci = read_image(input_file)
    outwcs = read_wcs(input_file)
    outsci = make_point_image(outsci, (500, 200), 40.0)
    inwcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=outwcs)
    driz.outsci = outsci

    driz.blot_image(inwcs)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("blot with point",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 16)
        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_blot_with_default(tmpdir):
    """
    Test do_blot with default grid image
    """
    output = str(tmpdir.join('output_blot_default.fits'))
    output_difference = str(tmpdir.join('difference_blot_default.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_blot_default.fits')

    outsci = read_image(input_file)
    outsci = make_grid_image(outsci, 64, 100.0)
    outwcs = read_wcs(input_file)
    inwcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=outwcs)
    driz.outsci = outsci

    driz.blot_image(inwcs)

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("blot with defaults",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 16)

        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_blot_with_lan3(tmpdir):
    """
    Test do_blot with lan3 grid image
    """
    output = str(tmpdir.join('output_blot_lan3.fits'))
    output_difference = str(tmpdir.join('difference_blot_lan3.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_blot_lan3.fits')

    outsci = read_image(input_file)
    outsci = make_grid_image(outsci, 64, 100.0)
    outwcs = read_wcs(input_file)
    inwcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=outwcs)
    driz.outsci = outsci

    driz.blot_image(inwcs, interp="lan3")

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("blot with lan3",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 16)
        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_blot_with_lan5(tmpdir):
    """
    Test do_blot with lan5 grid image
    """
    output = str(tmpdir.join('output_blot_lan5.fits'))
    output_difference = str(tmpdir.join('difference_blot_lan5.txt'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_blot_lan5.fits')

    outsci = read_image(input_file)
    outsci = make_grid_image(outsci, 64, 100.0)
    outwcs = read_wcs(input_file)
    inwcs = read_wcs(output_template)

    driz = drizzle.Drizzle(outwcs=outwcs)
    driz.outsci = outsci

    driz.blot_image(inwcs, interp="lan5")

    if ok:
        driz.write(output_template)
    else:
        driz.write(output)
        template_data = read_image(output_template)

        min_diff, med_diff, max_diff = centroid_statistics("blot with lan5",
                                                           output_difference,
                                                           driz.outsci,
                                                           template_data, 20.0, 16)
        assert med_diff < 1.0e-6
        assert max_diff < 1.0e-5


def test_context_planes():
    """Reproduce error seen in issue #50"""
    shape = [10, 10]
    outwcs = wcs.WCS()
    outwcs.pixel_shape = shape
    driz = drizzle.Drizzle(outwcs=outwcs)

    image = np.ones(shape)
    inwcs = wcs.WCS()
    inwcs.pixel_shape = shape

    for i in range(32):
        driz.add_image(image, inwcs)
    assert driz.outcon.shape == (1, 10, 10)

    driz.add_image(image, inwcs)
    assert driz.outcon.shape == (2, 10, 10)



import numpy as np


def calc_pixmap(first_wcs, second_wcs):
    """
    Calculate a mapping between the pixels of two images.

    Parameters
    ----------

    first_wcs : wcs
        A WCS object representing the coordinate system you are
        converting from

    seond_wcs : wcs
        A WCS object representing the coordinate system you are
        converting to

    Returns
    -------

    A three dimensional array representing the transformation between
    the two. The last dimension is of length two and contains the x and
    y coordinates of a pixel center, repectively. The other two coordinates
    correspond to the two coordinates of the image the first WCS is from.
    """

    first_naxis1, first_naxis2 = first_wcs.pixel_shape

    # We add one to the pixel co-ordinates before the transformation and subtract
    # it afterwards because wcs co-ordinates are one based, while pixel co-ordinates
    # are zero based, The result is the final values in pixmap give a tranformation
    # between the pixel co-ordinates in the first image to pixel co-ordinates in the
    # co-ordinate system of the second.

    one = np.ones(2, dtype='float64')
    idxmap = np.indices((first_naxis1, first_naxis2), dtype='float64')
    idxmap = idxmap.transpose() + one

    idxmap = idxmap.reshape(first_naxis2 * first_naxis1, 2)

    worldmap = first_wcs.all_pix2world(idxmap, 1)

    if second_wcs.sip is None:
        pixmap = second_wcs.wcs_world2pix(worldmap, 1)
    else:
        pixmap = second_wcs.all_world2pix(worldmap, 1)

    pixmap = pixmap.reshape(first_naxis2, first_naxis1, 2)
    pixmap = pixmap - one

    return pixmap



"""
The `drizzle` module defines the `Drizzle` class, for combining input
images into a single output image using the drizzle algorithm.
"""
import os
import os.path

import numpy as np
from astropy import wcs
from astropy.io import fits

from . import util
from . import doblot
from . import dodrizzle


class Drizzle(object):
    """
    Combine images using the drizzle algorithm
    """
    def __init__(self, infile="", outwcs=None,
                 wt_scl="exptime", pixfrac=1.0, kernel="square",
                 fillval="INDEF"):
        """
        Create a new Drizzle output object and set the drizzle parameters.

        All parameters are optional, but either infile or outwcs must be supplied.
        If infile initializes the object from a file written after a
        previous run of drizzle. Results from the previous run will be combined
        with new results. The value passed in outwcs will be ignored. If infile is
        not set, outwcs will be used to initilize a new run of drizzle.

        Parameters
        ----------

        infile : str, optional
            A fits file containing results from a previous run. The three
            extensions SCI, WHT, and CTX contain the combined image, total counts
            and image id bitmap, repectively. The WCS of the combined image is
            also read from the SCI extension.

        outwcs : wcs, optional
            The world coordinate system (WCS) of the combined image. This
            parameter must be present if no input file is given and is ignored if
            one is.

        wt_scl : str, optional
            How each input image should be scaled. The choices are `exptime`
            which scales each image by its exposure time, `expsq` which scales
            each image by the exposure time squared, or an empty string, which
            allows each input image to be scaled individually.

        pixfrac : float, optional
            The fraction of a pixel that the pixel flux is confined to. The
            default value of 1 has the pixel flux evenly spread across the image.
            A value of 0.5 confines it to half a pixel in the linear dimension,
            so the flux is confined to a quarter of the pixel area when the square
            kernel is used.

        kernel : str, optional
            The name of the kernel used to combine the inputs. The choice of
            kernel controls the distribution of flux over the kernel. The kernel
            names are: "square", "gaussian", "point", "tophat", "turbo", "lanczos2",
            and "lanczos3". The square kernel is the default.

        fillval : str, otional
            The value a pixel is set to in the output if the input image does
            not overlap it. The default value of INDEF does not set a value.
        """

        # Initialize the object fields

        self.outsci = None
        self.outwht = None
        self.outcon = None

        self.outexptime = 0.0
        self.uniqid = 0

        self.outwcs = outwcs
        self.wt_scl = wt_scl
        self.kernel = kernel
        self.fillval = fillval
        self.pixfrac = float(pixfrac)

        self.sciext = "SCI"
        self.whtext = "WHT"
        self.ctxext = "CTX"

        out_units = "cps"

        if not util.is_blank(infile):
            if os.path.exists(infile):
                handle = fits.open(infile)

                # Read parameters from image header
                self.outexptime = util.get_keyword(handle, "DRIZEXPT", default=0.0)
                self.uniqid = util.get_keyword(handle, "NDRIZIM", default=0)

                self.sciext = util.get_keyword(handle, "DRIZOUDA", default="SCI")
                self.whtext = util.get_keyword(handle, "DRIZOUWE", default="WHT")
                self.ctxext = util.get_keyword(handle, "DRIZOUCO", default="CTX")

                self.wt_scl = util.get_keyword(handle, "DRIZWTSC", default=wt_scl)
                self.kernel = util.get_keyword(handle, "DRIZKERN", default=kernel)
                self.fillval = util.get_keyword(handle, "DRIZFVAL", default=fillval)
                self.pixfrac = float(util.get_keyword(handle,
                                     "DRIZPIXF", default=pixfrac))

                out_units = util.get_keyword(handle, "DRIZOUUN", default="cps")

                try:
                    hdu = handle[self.sciext]
                    self.outsci = hdu.data.copy().astype(np.float32)
                    self.outwcs = wcs.WCS(hdu.header, fobj=handle)
                except KeyError:
                    pass

                try:
                    hdu = handle[self.whtext]
                    self.outwht = hdu.data.copy().astype(np.float32)
                except KeyError:
                    pass

                try:
                    hdu = handle[self.ctxext]
                    self.outcon = hdu.data.copy().astype(np.int32)
                    if self.outcon.ndim == 2:
                        self.outcon = np.reshape(self.outcon, (1,
                                                 self.outcon.shape[0],
                                                 self.outcon.shape[1]))

                    elif self.outcon.ndim == 3:
                        pass

                    else:
                        msg = ("Drizzle context image has wrong dimensions: " +
                               infile)
                        raise ValueError(msg)

                except KeyError:
                    pass

                handle.close()

        # Check field values

        if self.outwcs:
            util.set_pscale(self.outwcs)
        else:
            raise ValueError("Either an existing file or wcs must be supplied to Drizzle")

        if util.is_blank(self.wt_scl):
            self.wt_scl = ''
        elif self.wt_scl != "exptime" and self.wt_scl != "expsq":
            raise ValueError("Illegal value for wt_scl: %s" % out_units)

        if out_units == "counts":
            np.divide(self.outsci, self.outexptime, self.outsci)
        elif out_units != "cps":
            raise ValueError("Illegal value for wt_scl: %s" % out_units)

        # Initialize images if not read from a file
        outwcs_naxis1, outwcs_naxis2 = self.outwcs.pixel_shape
        if self.outsci is None:
            self.outsci = np.zeros(self.outwcs.pixel_shape[::-1],
                                   dtype=np.float32)

        if self.outwht is None:
            self.outwht = np.zeros(self.outwcs.pixel_shape[::-1],
                                   dtype=np.float32)
        if self.outcon is None:
            self.outcon = np.zeros((1, outwcs_naxis2, outwcs_naxis1),
                                   dtype=np.int32)

    def add_fits_file(self, infile, inweight="",
                      xmin=0, xmax=0, ymin=0, ymax=0,
                      unitkey="", expkey="", wt_scl=1.0):
        """
        Combine a fits file with the output drizzled image.

        Parameters
        ----------

        infile : str
            The name of the fits file, possibly including an extension.

        inweight : str, otional
            The name of a file containing a pixel by pixel weighting
            of the input data. If it is not set, an array will be generated
            where all values are set to one.

        xmin : float, otional
            This and the following three parameters set a bounding rectangle
            on the output image. Only pixels on the output image inside this
            rectangle will have their flux updated. Xmin sets the minimum value
            of the x dimension. The x dimension is the dimension that varies
            quickest on the image. If the value is zero or less, no minimum will
            be set in the x dimension. All four parameters are zero based,
            counting starts at zero.

        xmax : float, otional
            Sets the maximum value of the x dimension on the bounding box
            of the ouput image. If the value is zero or less, no maximum will
            be set in the x dimension.

        ymin : float, optional
            Sets the minimum value in the y dimension on the bounding box. The
            y dimension varies less rapidly than the x and represents the line
            index on the output image. If the value is zero or less, no minimum
            will be set in the y dimension.

        ymax : float, optional
            Sets the maximum value in the y dimension. If the value is zero or
            less, no maximum will be set in the y dimension.

        unitkey : string, optional
            The name of the header keyword containing the image units. The
            units can either be "counts" or "cps" (counts per second.) If it is
            left blank, the value is assumed to be "cps." If the value is counts,
            before using the input image it is scaled by dividing it by the
            exposure time.

        expkey : string, optional
            The name of the header keyword containing the exposure time. The
            exposure time is used to scale the image if the units are counts and
            to scale the image weighting if the drizzle was initialized with
            wt_scl equal to "exptime" or "expsq." If the value of this parameter
            is blank, the exposure time is set to one, implying no scaling.

        wt_scl : float, optional
            If drizzle was initialized with wt_scl left blank, this value will
            set a scaling factor for the pixel weighting. If drizzle was
            initialized with wt_scl set to "exptime" or "expsq", the exposure time
            will be used to set the weight scaling and the value of this parameter
            will be ignored.
        """

        insci = None
        inwht = None

        if not util.is_blank(infile):
            fileroot, extn = util.parse_filename(infile)

            if os.path.exists(fileroot):
                handle = fits.open(fileroot)
                hdu = util.get_extn(handle, extn=extn)

                if hdu is not None:
                    insci = hdu.data
                    inwcs = wcs.WCS(header=hdu.header)
                    insci = hdu.data.copy()
                handle.close()

        if insci is None:
            raise ValueError("Drizzle cannot find input file: %s" % infile)

        if not util.is_blank(inweight):
            fileroot, extn = util.parse_filename(inweight)

            if os.path.exists(fileroot):
                handle = fits.open(fileroot)
                hdu = util.get_extn(handle, extn=extn)

                if hdu is not None:
                    inwht = hdu.data.copy()
                handle.close()

        in_units = util.get_keyword(fileroot, unitkey, "cps")
        expin = util.get_keyword(fileroot, expkey, 1.0)

        self.add_image(insci, inwcs, inwht=inwht,
                       xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                       expin=expin, in_units=in_units, wt_scl=wt_scl)

    def add_image(self, insci, inwcs, inwht=None,
                  xmin=0, xmax=0, ymin=0, ymax=0,
                  expin=1.0, in_units="cps", wt_scl=1.0):
        """
        Combine an input image with the output drizzled image.

        Instead of reading the parameters from a fits file, you can set
        them by calling this lower level method. `Add_fits_file` calls
        this method after doing its setup.

        Parameters
        ----------

        insci : array
            A 2d numpy array containing the input image to be drizzled.
            it is an error to not supply an image.

        inwcs : wcs
            The world coordinate system of the input image. This is
            used to convert the pixels to the output coordinate system.

        inwht : array, optional
            A 2d numpy array containing the pixel by pixel weighting.
            Must have the same dimenstions as insci. If none is supplied,
            the weghting is set to one.

        xmin : float, optional
            This and the following three parameters set a bounding rectangle
            on the output image. Only pixels on the output image inside this
            rectangle will have their flux updated. Xmin sets the minimum value
            of the x dimension. The x dimension is the dimension that varies
            quickest on the image. If the value is zero or less, no minimum will
            be set in the x dimension. All four parameters are zero based,
            counting starts at zero.

        xmax : float, optional
            Sets the maximum value of the x dimension on the bounding box
            of the ouput image. If the value is zero or less, no maximum will
            be set in the x dimension.

        ymin : float, optional
            Sets the minimum value in the y dimension on the bounding box. The
            y dimension varies less rapidly than the x and represents the line
            index on the output image. If the value is zero or less, no minimum
            will be set in the y dimension.

        ymax : float, optional
            Sets the maximum value in the y dimension. If the value is zero or
            less, no maximum will be set in the y dimension.

        expin : float, optional
            The exposure time of the input image, a positive number. The
            exposure time is used to scale the image if the units are counts and
            to scale the image weighting if the drizzle was initialized with
            wt_scl equal to "exptime" or "expsq."

        in_units : str, optional
            The units of the input image. The units can either be "counts"
            or "cps" (counts per second.) If the value is counts, before using
            the input image it is scaled by dividing it by the exposure time.

        wt_scl : float, optional
            If drizzle was initialized with wt_scl left blank, this value will
            set a scaling factor for the pixel weighting. If drizzle was
            initialized with wt_scl set to "exptime" or "expsq", the exposure time
            will be used to set the weight scaling and the value of this parameter
            will be ignored.
        """

        insci = insci.astype(np.float32)
        util.set_pscale(inwcs)

        if inwht is None:
            inwht = np.ones(insci.shape, dtype=insci.dtype)
        else:
            inwht = inwht.astype(np.float32)

        if self.wt_scl == "exptime":
            wt_scl = expin
        elif self.wt_scl == "expsq":
            wt_scl = expin * expin

        self.increment_id()
        self.outexptime += expin

        dodrizzle.dodrizzle(insci, inwcs, inwht, self.outwcs,
                            self.outsci, self.outwht, self.outcon,
                            expin, in_units, wt_scl,
                            wcslin_pscale=inwcs.pscale, uniqid=self.uniqid,
                            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax,
                            pixfrac=self.pixfrac, kernel=self.kernel,
                            fillval=self.fillval)

    def blot_fits_file(self, infile, interp='poly5', sinscl=1.0):
        """
        Resample the output using another image's world coordinate system.

        Parameters
        ----------

        infile : str
            The name of the fits file containing the world coordinate
            system that the output file will be resampled to. The name may
            possibly include an extension.

        interp : str, optional
            The type of interpolation used in the resampling. The
            possible values are "nearest" (nearest neighbor interpolation),
            "linear" (bilinear interpolation), "poly3" (cubic polynomial
            interpolation), "poly5" (quintic polynomial interpolation),
            "sinc" (sinc interpolation), "lan3" (3rd order Lanczos
            interpolation), and "lan5" (5th order Lanczos interpolation).

        sincscl : float, optional
            The scaling factor for sinc interpolation.
        """
        blotwcs = None

        fileroot, extn = util.parse_filename(infile)

        if os.path.exists(fileroot):
            handle = fits.open(fileroot)
            hdu = util.get_extn(handle, extn=extn)

            if hdu is not None:
                blotwcs = wcs.WCS(header=hdu.header)
            handle.close()

        if not blotwcs:
            raise ValueError("Drizzle did not get a blot reference image")

        self.blot_image(blotwcs, interp=interp, sinscl=sinscl)

    def blot_image(self, blotwcs, interp='poly5', sinscl=1.0):
        """
        Resample the output image using an input world coordinate system.

        Parameters
        ----------

        blotwcs : wcs
            The world coordinate system to resample on.

        interp : str, optional
            The type of interpolation used in the resampling. The
            possible values are "nearest" (nearest neighbor interpolation),
            "linear" (bilinear interpolation), "poly3" (cubic polynomial
            interpolation), "poly5" (quintic polynomial interpolation),
            "sinc" (sinc interpolation), "lan3" (3rd order Lanczos
            interpolation), and "lan5" (5th order Lanczos interpolation).

        sincscl : float, optional
            The scaling factor for sinc interpolation.
        """

        util.set_pscale(blotwcs)
        self.outsci = doblot.doblot(self.outsci, self.outwcs, blotwcs,
                                    1.0, interp=interp, sinscl=sinscl)

        self.outwcs = blotwcs

    def increment_id(self):
        """
        Increment the id count and add a plane to the context image if needed

        Drizzle tracks which input images contribute to the output image
        by setting a bit in the corresponding pixel in the context image.
        The uniqid indicates which bit. So it must be incremented each time
        a new image is added. Each plane in the context image can hold 32 bits,
        so after each 32 images, a new plane is added to the context.
        """

        # Compute what plane of the context image this input would
        # correspond to:
        planeid = int(self.uniqid / 32)

        # Add a new plane to the context image if planeid overflows

        if self.outcon.shape[0] == planeid:
            plane = np.zeros_like(self.outcon[0])
            self.outcon = np.append(self.outcon, [plane], axis=0)

        # Increment the id
        self.uniqid += 1

    def write(self, outfile, out_units="cps", outheader=None):
        """
        Write the output from a set of drizzled images to a file.

        The output file will contain three extensions. The "SCI" extension
        contains the resulting image. The "WHT" extension contains the
        combined weights. The "CTX" extension is a bit map. The nth bit
        is set to one if the nth input image contributed non-zero flux
        to the output image. The "CTX" image is three dimensionsional
        to account for the possibility that there are more than 32 input
        images.

        Parameters
        ----------

        outfile : str
            The name of the output file. If the file already exists,
            the old file is deleted after writing the new file.

        out_units : str, optional
            The units of the output image, either `counts` or `cps`
            (counts per second.) If the units are counts, the resulting
            image will be multiplied by the computed exposure time.

        outheader : header, optional
            A fits header containing cards to be added to the primary
            header of the output image.
        """

        if out_units != "counts" and out_units != "cps":
            raise ValueError("Illegal value for out_units: %s" % str(out_units))

        # Write the WCS to the output image

        handle = self.outwcs.to_fits()
        phdu = handle[0]

        # Write the class fields to the primary header
        phdu.header['DRIZOUDA'] = \
            (self.sciext, 'Drizzle, output data image')
        phdu.header['DRIZOUWE'] = \
            (self.whtext, 'Drizzle, output weighting image')
        phdu.header['DRIZOUCO'] = \
            (self.ctxext, 'Drizzle, output context image')
        phdu.header['DRIZWTSC'] = \
            (self.wt_scl, 'Drizzle, weighting factor for input image')
        phdu.header['DRIZKERN'] = \
            (self.kernel, 'Drizzle, form of weight distribution kernel')
        phdu.header['DRIZPIXF'] = \
            (self.pixfrac, 'Drizzle, linear size of drop')
        phdu.header['DRIZFVAL'] = \
            (self.fillval, 'Drizzle, fill value for zero weight output pix')
        phdu.header['DRIZOUUN'] = \
            (out_units, 'Drizzle, units of output image - counts or cps')

        # Update header keyword NDRIZIM to keep track of how many images have
        # been combined in this product so far
        phdu.header['NDRIZIM'] = (self.uniqid, 'Drizzle, number of images')

        # Update header of output image with exptime used to scale the output data
        # if out_units is not counts, this will simply be a value of 1.0
        # the keyword 'exptime' will always contain the total exposure time
        # of all input image regardless of the output units

        phdu.header['EXPTIME'] = \
            (self.outexptime, 'Drizzle, total exposure time')

        outexptime = 1.0
        if out_units == 'counts':
            np.multiply(self.outsci, self.outexptime, self.outsci)
            outexptime = self.outexptime
        phdu.header['DRIZEXPT'] = (outexptime, 'Drizzle, exposure time scaling factor')

        # Copy the optional header to the primary header

        if outheader:
            phdu.header.extend(outheader, unique=True)

        # Add three extensions containing, the drizzled output image,
        # the total counts, and the context bitmap, in that order

        extheader = self.outwcs.to_header()

        ehdu = fits.ImageHDU()
        ehdu.data = self.outsci
        ehdu.header['EXTNAME'] = (self.sciext, 'Extension name')
        ehdu.header['EXTVER'] = (1, 'Extension version')
        ehdu.header.extend(extheader, unique=True)
        handle.append(ehdu)

        whdu = fits.ImageHDU()
        whdu.data = self.outwht
        whdu.header['EXTNAME'] = (self.whtext, 'Extension name')
        whdu.header['EXTVER'] = (1, 'Extension version')
        whdu.header.extend(extheader, unique=True)
        handle.append(whdu)

        xhdu = fits.ImageHDU()
        xhdu.data = self.outcon
        xhdu.header['EXTNAME'] = (self.ctxext, 'Extension name')
        xhdu.header['EXTVER'] = (1, 'Extension version')
        xhdu.header.extend(extheader, unique=True)
        handle.append(xhdu)

        handle.writeto(outfile, overwrite=True)
        handle.close()



import os

import pytest
import numpy as np
from astropy import wcs
from astropy.io import fits

from drizzle import drizzle
from drizzle import util


TEST_DIR = os.path.abspath(os.path.dirname(__file__))
DATA_DIR = os.path.join(TEST_DIR, 'data')


def read_header(filename):
    """
    Read the primary header from a fits file
    """
    fileroot, extn = util.parse_filename(os.path.join(DATA_DIR, filename))
    with fits.open(fileroot) as hdulist:
        header = hdulist[0].header
    return header


def read_image(filename):
    """
    Read the image from a fits file
    """
    fileroot, extn = util.parse_filename(os.path.join(DATA_DIR, filename))
    with fits.open(fileroot, memmap=False) as hdulist:
        data = hdulist[1].data.copy()
    return data


def read_wcs(filename):
    """
    Read the wcs of a fits file
    """
    fileroot, extn = util.parse_filename(os.path.join(DATA_DIR, filename))
    with fits.open(fileroot) as hdulist:
        the_wcs = wcs.WCS(hdulist[1].header)
    return the_wcs


@pytest.fixture
def run_drizzle_reference_square_points(tmpdir):
    """Create an empty drizzle image"""
    output_file = str(tmpdir.join('output_null_run.fits'))
    output_template = os.path.join(DATA_DIR, 'reference_square_point.fits')

    output_wcs = read_wcs(output_template)
    driz = drizzle.Drizzle(outwcs=output_wcs, wt_scl="expsq", pixfrac=0.5,
                           kernel="turbo", fillval="NaN")
    driz.write(output_file)

    return output_file


def test_null_run(run_drizzle_reference_square_points):
    output_file = run_drizzle_reference_square_points
    with fits.open(output_file) as hdulist:

        assert hdulist.index_of("SCI") == 1
        assert hdulist.index_of("WHT") == 2
        assert hdulist.index_of("CTX") == 3

        pheader = hdulist["PRIMARY"].header

    assert pheader['DRIZOUDA'] == 'SCI'
    assert pheader['DRIZOUWE'] == 'WHT'
    assert pheader['DRIZOUCO'] == 'CTX'
    assert pheader['DRIZWTSC'] == 'expsq'
    assert pheader['DRIZKERN'] == 'turbo'
    assert pheader['DRIZPIXF'] == 0.5
    assert pheader['DRIZFVAL'] == 'NaN'
    assert pheader['DRIZOUUN'] == 'cps'
    assert pheader['EXPTIME'] == 0.0
    assert pheader['DRIZEXPT'] == 1.0


def test_file_init(run_drizzle_reference_square_points):
    """
    Initialize drizzle object from a file
    """
    input_file = run_drizzle_reference_square_points

    driz = drizzle.Drizzle(infile=input_file)

    assert driz.outexptime == 1.0
    assert driz.wt_scl == 'expsq'
    assert driz.kernel == 'turbo'
    assert driz.pixfrac == 0.5
    assert driz.fillval == 'NaN'


@pytest.fixture
def add_header(tmpdir):
    """Add extra keywords read from the header"""
    output_file = str(tmpdir.join('output_add_header.fits'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits')
    output_template = os.path.join(DATA_DIR, 'reference_square_point.fits')

    driz = drizzle.Drizzle(infile=output_template)
    image = read_image(input_file)
    inwcs = read_wcs(input_file)
    driz.add_image(image, inwcs)

    header = fits.header.Header()
    header['ONEVAL'] = (1.0, 'test value')
    header['TWOVAL'] = (2.0, 'test value')

    driz.write(output_file, outheader=header)

    return output_file


def test_add_header(add_header):
    output_file = add_header
    header = read_header(output_file)
    assert header['ONEVAL'] == 1.0
    assert header['TWOVAL'] == 2.0
    assert header['DRIZKERN'] == 'square'


def test_add_file(add_header, tmpdir):
    """
    Add an image read from a file
    """
    test_file = add_header
    output_file = str(tmpdir.join('output_add_file.fits'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits[1]')
    output_template = os.path.join(DATA_DIR, 'reference_square_point.fits')

    driz = drizzle.Drizzle(infile=output_template)
    driz.add_fits_file(input_file)
    driz.write(output_file)

    output_image = read_image(output_file)
    test_image = read_image(test_file)
    diff_image = np.absolute(output_image - test_image)

    assert np.amax(diff_image) == 0.0


def test_blot_file(tmpdir):
    """
    Blot an image read from a file
    """
    output_file = str(tmpdir.join('output_blot_file.fits'))
    test_file = str(tmpdir.join('output_blot_image.fits'))

    input_file = os.path.join(DATA_DIR, 'j8bt06nyq_flt.fits[1]')
    output_template = os.path.join(DATA_DIR, 'reference_blot_image.fits')

    blotwcs = read_wcs(input_file)

    driz = drizzle.Drizzle(infile=output_template)
    driz.add_fits_file(input_file)
    driz.blot_image(blotwcs)
    driz.write(test_file)

    driz = drizzle.Drizzle(infile=output_template)
    driz.add_fits_file(input_file)
    driz.blot_fits_file(input_file)
    driz.write(output_file)

    output_image = read_image(output_file)
    test_image = read_image(test_file)
    diff_image = np.absolute(output_image - test_image)

    assert np.amax(diff_image) == 0.0



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import sys
from setuptools import setup, Extension

import numpy


def get_extensions():
    srcdir = os.path.join(os.path.dirname(__file__), 'src')
    cdriz_sources = ['cdrizzleapi.c',
                     'cdrizzleblot.c',
                     'cdrizzlebox.c',
                     'cdrizzlemap.c',
                     'cdrizzleutil.c',
                     os.path.join('tests', 'utest_cdrizzle.c')]
    sources = [os.path.join(srcdir, x) for x in cdriz_sources]

    cfg = {
        'include_dirs': [],
        'libraries': [],
        'define_macros': [],
    }
    cfg['include_dirs'].append(numpy.get_include())
    cfg['include_dirs'].append(srcdir)
    if sys.platform != 'win32':
        cfg['libraries'].append('m')
    if sys.platform == 'win32':
        cfg['define_macros'].append(('WIN32', None))
        cfg['define_macros'].append(('__STDC__', 1))
        cfg['define_macros'].append(('_CRT_SECURE_NO_WARNINGS', None))

    return [Extension(str('drizzle.cdrizzle'), sources, **cfg)]


setup(ext_modules=get_extensions())



"""
STScI Python compatable drizzle module
"""
import numpy as np

from . import util
from . import calc_pixmap
from . import cdrizzle


def dodrizzle(insci, input_wcs, inwht,
              output_wcs, outsci, outwht, outcon,
              expin, in_units, wt_scl,
              wcslin_pscale=1.0, uniqid=1,
              xmin=0, xmax=0, ymin=0, ymax=0,
              pixfrac=1.0, kernel='square', fillval="INDEF"):
    """
    Low level routine for performing 'drizzle' operation.on one image.

    The interface is compatible with STScI code. All images are Python
    ndarrays, instead of filenames. File handling (input and output) is
    performed by the calling routine.

    Parameters
    ----------

    insci : 2d array
        A 2d numpy array containing the input image to be drizzled.
        it is an error to not supply an image.

    input_wcs : 2d array
        The world coordinate system of the input image.

    inwht : 2d array
        A 2d numpy array containing the pixel by pixel weighting.
        Must have the same dimensions as insci. If none is supplied,
        the weghting is set to one.

    output_wcs : wcs
        The world coordinate system of the output image.

    outsci : 2d array
        A 2d numpy array containing the output image produced by
        drizzling. On the first call it should be set to zero.
        Subsequent calls it will hold the intermediate results

    outwht : 2d array
        A 2d numpy array containing the output counts. On the first
        call it should be set to zero. On subsequent calls it will
        hold the intermediate results.

    outcon : 2d or 3d array, optional
        A 2d or 3d numpy array holding a bitmap of which image was an input
        for each output pixel. Should be integer zero on first call.
        Subsequent calls hold intermediate results.

    expin : float
        The exposure time of the input image, a positive number. The
        exposure time is used to scale the image if the units are counts.

    in_units : str
        The units of the input image. The units can either be "counts"
        or "cps" (counts per second.)

    wt_scl : float
        A scaling factor applied to the pixel by pixel weighting.

    wcslin_pscale : float, optional
        The pixel scale of the input image. Conceptually, this is the
        linear dimension of a side of a pixel in the input image, but it
        is not limited to this and can be set to change how the drizzling
        algorithm operates.

    uniqid : int, optional
        The id number of the input image. Should be one the first time
        this function is called and incremented by one on each subsequent
        call.

    xmin : float, optional
        This and the following three parameters set a bounding rectangle
        on the input image. Only pixels on the input image inside this
        rectangle will have their flux added to the output image. Xmin
        sets the minimum value of the x dimension. The x dimension is the
        dimension that varies quickest on the image. If the value is zero,
        no minimum will be set in the x dimension. All four parameters are
        zero based, counting starts at zero.

    xmax : float, optional
        Sets the maximum value of the x dimension on the bounding box
        of the input image. If the value is zero, no maximum will
        be set in the x dimension, the full x dimension of the output
        image is the bounding box.

    ymin : float, optional
        Sets the minimum value in the y dimension on the bounding box. The
        y dimension varies less rapidly than the x and represents the line
        index on the input image. If the value is zero, no minimum  will be
        set in the y dimension.

    ymax : float, optional
        Sets the maximum value in the y dimension. If the value is zero, no
        maximum will be set in the y dimension,  the full x dimension
        of the output image is the bounding box.

    pixfrac : float, optional
        The fraction of a pixel that the pixel flux is confined to. The
        default value of 1 has the pixel flux evenly spread across the image.
        A value of 0.5 confines it to half a pixel in the linear dimension,
        so the flux is confined to a quarter of the pixel area when the square
        kernel is used.

    kernel: str, optional
        The name of the kernel used to combine the input. The choice of
        kernel controls the distribution of flux over the kernel. The kernel
        names are: "square", "gaussian", "point", "tophat", "turbo", "lanczos2",
        and "lanczos3". The square kernel is the default.

    fillval: str, optional
        The value a pixel is set to in the output if the input image does
        not overlap it. The default value of INDEF does not set a value.

    Returns
    -------
    A tuple with three values: a version string, the number of pixels
    on the input image that do not overlap the output image, and the
    number of complete lines on the input image that do not overlap the
    output input image.

    """

    # Ensure that the fillval parameter gets properly interpreted
    # for use with tdriz
    if util.is_blank(fillval):
        fillval = 'INDEF'
    else:
        fillval = str(fillval)

    if in_units == 'cps':
        expscale = 1.0
    else:
        expscale = expin

    # Add input weight image if it was not passed in

    if (insci.dtype > np.float32):
        insci = insci.astype(np.float32)

    if inwht is None:
        inwht = np.ones_like(insci)

    # Compute what plane of the context image this input would
    # correspond to:
    planeid = int((uniqid - 1) / 32)

    # Check if the context image has this many planes
    if outcon.ndim == 3:
        nplanes = outcon.shape[0]
    elif outcon.ndim == 2:
        nplanes = 1
    else:
        nplanes = 0

    if nplanes <= planeid:
        raise IndexError("Not enough planes in drizzle context image")

    # Alias context image to the requested plane if 3d
    if outcon.ndim == 3:
        outcon = outcon[planeid]

    pix_ratio = output_wcs.pscale / wcslin_pscale

    # Compute the mapping between the input and output pixel coordinates
    pixmap = calc_pixmap.calc_pixmap(input_wcs, output_wcs)

    #
    # Call 'drizzle' to perform image combination
    # This call to 'cdriz.tdriz' uses the new C syntax
    #
    _vers, nmiss, nskip = cdrizzle.tdriz(
        insci, inwht, pixmap, outsci, outwht, outcon,
        uniqid=uniqid, xmin=xmin, xmax=xmax,
        ymin=ymin, ymax=ymax, scale=pix_ratio, pixfrac=pixfrac,
        kernel=kernel, in_units=in_units, expscale=expscale,
        wtscale=wt_scl, fillstr=fillval)

    return _vers, nmiss, nskip
